// DCF.CPP    DCF77-Zeit fuer WEBIO von Elektor
//
// Hardware : Mini-Webio von Elektor mit 2 Ports je 8 bit Adr. 100h und 101h
//            LCD-Anzeige 2 Zeilen je 40 Zeichen          Adr. 102h und 103h        
//            Aus/Eingabe-Testboard 
//            DCF77-Modul                an Bit D7 Ein    Adr. 100h
//            Stop-Taste                 an Bit D2 Ein    Adr. 100h
//
// Software : Borland Turbo C++ V1.01 
//
// Author :   Diedrich Lamken Schaphuser Str. 2 28307 Bremen      05.01.2002
//            Stoerungen in DCF-Impulsen                          28.01.2002
//            Test                                                19.01.2004

// Quellen :  Elektor Hefte 5 bis 9 / 2001
//            D.Schwarzer BBS Winsen/Luhe
//            C.Kuehnel   Embedded Webserver mit Single-Chip Controller


#pragma option -1               // 80186 Code erzeugen

// Includes
//---------
#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>


// Definitionen
//------------- 
#define  HAL_INT  0xA1           // Interrupts
#define  PFE_INT  0xA2
#define  TCP_INT  0xAC
#define  DOS_INT  0x21

#define  port1adr 0x100          // Port 1         
#define  port2adr 0x101          // Port 2
#define  lcdcom   0x102          // LCD-Anzeige com
#define  lcddat   0x103          //             Daten


// globale Variablen
// -----------------

union    REGS  inregs;          // C-SoftInterrupt-Funktion
union    REGS  outregs;
struct   SREGS segregs;
				// DCF-Zeitinformationen
unsigned int   ptbword  = 0;    // 15Bit PTB-Meldungen
unsigned char  flags    = 0;    //  5Bit Ank.Schaltsek,
				//       Zone2, Zone1,
				//       Ank. MEZ/MESZ,
				//       Antenne
unsigned char  sekunde  = 0;    // nach dem Empfang 0
unsigned char  minute   = 0;    // ..von dcf77
unsigned char  stunde   = 0;    // ..von dcf77
unsigned char  tag      = 0;    // ..von dcf77
unsigned char  tagnr    = 0;    // ..von dcf77  0-So,1-Mo
unsigned char  monat    = 0;    // ..von dcf77
unsigned int   jahr     = 0;    // ..von dcf77
unsigned int   jahrt    = 2000;
	 char  *wotag[] =
         {"Sonntag",   "Montag", "Dienstag", "Mittwoch"
	 ,"Donnerstag","Freitag","Sonnabend","Sonntag"};
				// Benutzer SoftINT-Handler
volatile unsigned ms;           // msec - Zaehler
unsigned       old_es;          // Adr. alter Handler HIGH
unsigned       old_bx;          // Adr. alter Handler LOW


//   Prototypes
//   ----------
void get_time(void);
void set_time(void);
void lcd_write ( unsigned char x, unsigned char y, char *text );
void fix ( unsigned int zahl, char buffaus[10], unsigned char lenmax ) ;
void fix0( unsigned int zahl, char buffaus[10], unsigned char lenmax ) ;


// Funktionen des Timers
// ---------------------


void interrupt handler (void)
//             -------
// die Funktion wird vom abgelaufenen Timer jede ms
// aufgerufen und und erhht die Variable ms
{
   ms++;
}

void init_msCount (int timer)
//   ------------
// installiert den InterruptHandler, fragt die System-
// frequenz fuer Timer ab und initialisiert einen
// der Timer 0 oder 1
{
  // Interrupt-Handler fuer benutzten Timer setzen
  inregs.h.ah = 0x84;                       // SoftINT A1  Fkt 84
  inregs.x.cx = 1;                          // Anzahl new INT
  inregs.x.dx = 8+timer;                    // Timer 0 od.1
  segregs.es  = FP_SEG(handler);            // Handler-adr.high
  inregs.x.bx = FP_OFF(handler);            // Handler-adr.low
  int86x(HAL_INT, &inregs, &outregs, &segregs);
  old_es      = segregs.es;                 // alte Adresse high
  old_bx      = outregs.x.bx;               // alte Adresse low

  // Timertakt des Systems zur Berechnung der Wertes 1ms
  unsigned long tfreq;
  inregs.h.ah = 0x8A;                       // SoftINT A1  Fkt 8A
  inregs.h.al = 1;                          // Timerfrequenz
  int86x(HAL_INT, &inregs, &outregs, &segregs);
  tfreq       = outregs.x.dx;               // wegen 32Bit
  tfreq       = ( (tfreq<<16) + outregs.x.ax) / 500;

  // benutzten Timer voreinstellen
  inregs.h.ah = 0x85;                       // SoftINT A1  Fkt 85
  inregs.h.al = timer;                      // Timer 0 od. 1
  inregs.x.dx = 0x0003;                     // DauerLauf, TimerINT
  inregs.x.cx = tfreq;                      // Abbruchwert
  int86x(HAL_INT, &inregs, &outregs, &segregs);
}

void oldHandler(int timer)
//   ----------
// setzt am Ende des Programms die Adressen des
// urspruenglichen SoftINT-Handlers wieder ein.
{
  inregs.h.ah = 0x84;                       // SoftINT A1  Fkt 84
  inregs.x.dx = 8 + timer;                  // Timer 0 od.1
  segregs.es  = old_es;                     // Handler-adr.high
  inregs.x.bx = old_bx;                     // Handler-adr.low
  int86x(HAL_INT, &inregs, &outregs, &segregs);
}

void start_timer(int timer)
//   ----------
// startet den benutzten timer, mit den gesetzten Werten.
// loescht zuvor den msec-Zaehler.
// stop_timer()  haelt ihn wieder an.
{
  ms          = 0;                          // msec-Zaehler auf 0
  inregs.h.ah = 0x86;                       // SoftINT A1  Fkt 86
  inregs.h.al = timer;
  int86x(HAL_INT, &inregs, &outregs, &segregs);
}

void stop_timer(int timer)
//   ---------
{
  inregs.h.ah = 0x87;                       // SoftINT A1  Fkt 87
  inregs.h.al = timer;
  int86x(HAL_INT, &inregs, &outregs, &segregs);
}


void sleep(int msec)
//   -----
// veranlasst eine Wartezeit, die in msec angegeben wird.
{
   inregs.h.ah = 0x09;                      // SoftINT: ACh, Fkt: 09h
   inregs.x.bx = msec;
   int86x(TCP_INT, &inregs, &outregs, &segregs);
}


// Funktionen des Eingabeports
// ---------------------------

void port_init (void)
//   -------
// funktioniert beim Elektor-Webio nicht wie gewuenscht, 
// deshalb  ALE 1  und PCS 1 in AUTOEXEC.BAT
//
// -setzt die Leitungen AD0 .. AD7 auf bidirektionalen Betrieb
// -sperrt die Leitung ALE (wird 1)
// -setzt  das Signal #PCS6 (A2), so wird das Tor unter den
//         Adressen 600h-6FFh ansprechbar
//
//  geaendert auf PCS 1 und Adresse 101h
//  fuehrt aber zu Problemen beim Einlesen von D7 an Adr. 101h
{
   inregs.h.ah = 0x80;                       // SoftINT A2  Fkt 80
   inregs.h.al = 0x00;                       // sperren von ALE
   inregs.x.dx = 0xFF;                       // AD-Pins bidirektional
   int86x(PFE_INT, &inregs, &outregs, &segregs);

   inregs.h.ah = 0x83;                       // SoftINT A2  Fkt 83
   inregs.x.dx = 0x02;                       // #PCS1 0000 0010 => ab 100h

   int86x(PFE_INT, &inregs, &outregs, &segregs);
}

unsigned char port_read (void)
//            -------
// Liest die Zustaende an den Eingabepins I0-I7
{
  inregs.h.ah = 0x80;                       // SoftINT A1  Fkt 80

  inregs.x.di = port2adr;                   // Adresse 101h ausgeben                                        // geaendert Lamken
  inregs.x.bx = 0xFFFF;                     // wAND=0xFFFF
  inregs.x.cx = 0x0000;                     // wXOR=0x0000
  int86x(HAL_INT, &inregs, &outregs, &segregs);
  return outregs.h.al;
}

unsigned char port_in ( int portadr )
//            -------
// Liest die Zustaende an den Eingabepins I0-I7
{
  inregs.h.ah = 0x80;                       // SoftINT A1  Fkt 80

  inregs.x.di = portadr;                    // an Adresse ausgeben                                        // geaendert Lamken
  inregs.x.bx = 0xFFFF;                     // wAND=0xFFFF
  inregs.x.cx = 0x0000;                     // wXOR=0x0000
  int86x(HAL_INT, &inregs, &outregs, &segregs);
  return outregs.h.al;
}

void port_out( int portadr, unsigned char ausgabe)
//   --------
// schreibt in die Ausgabepins O0-O7 den Wert 'ausgabe'
{
   inregs.h.ah = 0x81;                      // SoftINT: A1h, Fkt: 81h
   inregs.x.di = portadr;                   // Adresse ausgeben
   inregs.h.dh = 0;
   inregs.h.dl = ausgabe;
   inregs.x.bx = 0xFFFF;                    // wAND=0xFFFF
   inregs.x.cx = 0x0000;                    // wXOR=0x0000
   int86x(HAL_INT, &inregs, &outregs, &segregs);
}


// Funktionen der DOS-Uhr
// ----------------------

void set_time (void)
//   ---------
// nachdem die Zeit der Funkuhr empfangen wurde, uebertragen
// sie diese beiden Funktionen in den Zeitspeicher des SC12
// Betriebsystems. Hier wird die Zeit auf Quarzgenauigkeit
// fortgeschrieben (bis zu einem erneuten Empfang).
{
  inregs.h.ah = 0x2D;                       // SoftINT 21  Fkt 2D
  inregs.h.ch = stunde;
  inregs.h.cl = minute;
  inregs.h.dh = sekunde+1;                  // Korrektur zu Wecker
  int86x(DOS_INT, &inregs, &outregs, &segregs);
}


void get_time(void)
//   --------
// holt die System-Zeit des IPC
{
   inregs.h.ah = 0x2C;
   int86x(DOS_INT, &inregs, &outregs, &segregs);
   stunde      = outregs.h.ch;
   minute      = outregs.h.cl;
   sekunde     = outregs.h.dh;
// printf("Die aktuelle Zeit ist: %2d:%2d:%2d", stunde, minute, sekunde);
}


void set_date (void)
//   ----------
{
   inregs.h.ah = 0x2B;                       // SoftINT 21  Fkt 2B
   inregs.x.cx = jahr;
   inregs.h.dl = tag;
   inregs.h.dh = monat;
   int86x(DOS_INT, &inregs, &outregs, &segregs);
}


void get_date(void)
//   --------
// holt das System-Datum des IPC
{
   inregs.h.ah = 0x2A;
   int86x(DOS_INT, &inregs, &outregs, &segregs);
   jahr        = outregs.x.cx;
   monat       = outregs.h.dh;
   tag         = outregs.h.dl;
   tagnr       = outregs.h.al;
// printf("Das aktuelle Datum ist: %4d-%2d-%2d", jahr, monat, tag);
}

void show_date(void)
//   ---------
// gibt System-Datum und -Zeit des IPC aus
{
   get_date(); get_time();

   printf("\n %s, der %0.2d.%0.2d.%0.4d", wotag[tagnr], tag, monat, jahr);
   printf(" um %0.2d:%0.2d:%0.2d", stunde, minute, sekunde);
}

void lcd_date(void)
//   --------
// gibt System-Datum und -Zeit des IPC auf LCD-Anzeige aus
{
   char zeile1[41], zeile2[41];
   sprintf(zeile2," %s, der %2d.%0.2d.%4d um %2d:%0.2d:%0.2d",
      wotag[tagnr], tag, monat, jahr, stunde, minute, sekunde);
   lcd_write(0,0,zeile2);
}

// Funktionen der DCF-Zeit
// -----------------------
// Diese Variablen koennen entfallen, wenn in der Funktion
// statistik() die Varianzgroessen nicht erfasst werden sollen

int maxmin = 0, maxmax = 0;                 // globale Variable
int minmin = 0, minmax = 0;

void statistik (int on)
//   ---------
// die Funktion stellt die groessten und die kleinsten Werte
// fuer on=0 und on=1 fuer die Bildschirmausgabe zusammen.
// Sie kann entfallen, wenn diese Werte ohne Belang sind.
// Aufruf aus getDCF();
{
  static int ma = 0, mi = 0;
  if (on > 140) {                                     // 1-Signal
     if ( ma==0 )   { maxmin = maxmax = on; ma = 1; }
     if ( on<maxmin ) maxmin = on;
     if ( on>maxmax ) maxmax = on; }
  else  {                                             // 0-Signal
     if ( mi==0 )   { minmin = minmax = on; mi = 1; }
     if ( on<minmin ) minmin = on;
     if ( on>minmax ) minmax = on; }
}


void printDCF(void)
//   --------
// gibt die zusammengestellten DCF-Zeitinformationen zum
// Bildschirm aus.
{
  char *R    = " R       : es sendet die ";
  char *A1   = " A1      : Wechsel MEZ/MESZ - ";
  char *Z1   = " Z1      : Zeitzone - ";
  char *Z2   = " Z2      : Zeitzone - ";
  char *A2   = " A2      : Schaltsekunde - ";
  char *txt  = "zur vollen Stunde";
  char *nein = "nein";
  char *mez  = "MEZ";
  char *mesz = "MESZ";
  int  maske = 0x4000;

  printf("\n\n PTB-Code: ");
  while (maske) { if ( ptbword&maske ) printf("1");
		  else                 printf("0");
		  maske = maske>>1;
		}
  if ( flags & 0x01 )                printf("\n%sReserveantenne",R);
       else                          printf("\n%sBetriebsantenne",R);
  if ( flags & 0x02 )                printf("\n%s%s",A1,txt);
       else                          printf("\n%s%s",A1,nein);
  if ( flags & 0x04 )                printf("\n%s%s",Z1,mesz);
       else                          printf("\n%s%s",Z1,mez);
  if ( flags & 0x08 )                printf("\n%s%s",Z2,mez);
       else                          printf("\n%s%s",Z2,mesz);
  if ( flags & 0x10 )                printf("\n%s%s",A2,txt);
       else                          printf("\n%s%s\n",A2,nein);

  printf("\n Datum: %s %0.2d:%0.2d:%0.4d",
          wotag[tagnr], tag, monat, jahr);
  printf(" Zeit:  %0.2d:%0.2d:%0.2d",
	  stunde, minute, sekunde);
}


int getDCF (unsigned on)
//  ------
// Die Funktion stellt fest, ob es sich bei dem uebergebenen
// Wert in ms um ein 0-Signal (100ms) oder um ein 1-Signal
// (200ms)des DCF-Empfaengers handelt.
// Werte ausserhalb der angegebenen Toleranzbereiche sprechen fuer
// ein Stoersignal. In diesem Fall wird -1 zurueckgegeben.
// Die Funktion statistik haelt die groessten und kleinsten Werte
// von on=0 und on=1 fest.
{
  statistik(on);                           //  alte Werte von Schwarzer
  if (on >=  60 && on <= 110) return(0);   //  80...110 msec
  if (on >= 160 && on <= 205) return(1);   // 175...205 msec
  return (-1);
}


unsigned BCDtoHEX(unsigned bcd)
//       --------
// die DCF-Zeitangaben werden im BCD-Format bermittelt
// DOS-Uhren benutzen das HEX-Format. Die Funktion erzeugt
// aus einer maximal 4-stelligen gepackten BDC-Zahl
// eine Hexzahl.
{
  bcd= (( bcd&0xF000 ) >>12 ) * 1000 +
       (( bcd&0x0F00 ) >> 8 ) *  100 +
       (( bcd&0x00F0 ) >> 4 ) *   10 +
	( bcd&0x000F );
  return(bcd);
}


int decodeDCF (int s, int bit)
//  ---------
// die Funktion stellt die empfangenen DCF-Bits zu den
// enthaltenen Informationspaketen Min, Std, Tag,
// Monat, Jahr, usw .. zusammen und kontrolliert deren
// Pruefsummen.
// Rueckgaben:  0 - alles ok
//             -1 - uebergebenes Bit war weder 0 noch 1
//             -2 - Sekunde 20 Zeitstartinformation falsch
//             -3 - Pruefsumme Minuten falsch
//             -4 - Pruefsumme Stunden falsch
//             -5 - Pruefsumme Datum falsch
//             -6 - Sekunde groesser als 58
{
  static unsigned pbit=0;

  if ( bit != 1 && bit != 0) return(-1);

  pbit = pbit^bit;         // Pruefsumme berechnen mit XOR

                           // PTB-Wort und flags
  if      (s>= 0 && s<=14) ptbword = ( ptbword>>1 ) + ( bit<<14 );
  else if (s>=15 && s<=19) flags = ( flags>>1 ) + ( bit<<4  );

			   // Bit 20: Start Zeitinfo, immer 1
  else if (s==20)          { if(!bit) return(-2); pbit = 0;}
  else if (s>=21 && s<=27) minute  = ( minute>>1 ) + ( bit<<6 );
  else if (s==28)          { if(pbit) return(-3);
                           minute  = BCDtoHEX(minute); }

  else if (s>=29 && s<=34) stunde  = ( stunde>>1 ) + ( bit<<5 );
  else if (s==35)          { if(pbit) return(-4);
                           stunde  = BCDtoHEX(stunde); }

  else if (s>=36 && s<=41) tag     = ( tag>>1   ) + ( bit<<5 );
  else if (s>=42 && s<=44) tagnr   = ( tagnr>>1 ) + ( bit<<2 );
  else if (s>=45 && s<=49) monat   = ( monat>>1 ) + ( bit<<4 );
  else if (s>=50 && s<=57) jahr    = ( jahr>>1  ) + ( bit<<7 );
  else if (s==58)          { if(pbit) return(-5);
                           tag     = BCDtoHEX(tag);
			   monat   = BCDtoHEX(monat);
			   jahr    = BCDtoHEX(jahr)+jahrt; }
  else    return(-6);      // Sek > 58
	  return( 0);      // alles i.o.
}


// Funktionen der LCD-Anzeige
// --------------------------

void lcd_wait()
//   ----------
// wartet bis LCD-Anzeige fertig
{  do {}
   while (( volatile unsigned char) ( port_in(lcdcom) ) > 127 );
}

void lcd_init()
//   ----------
// initialisiert LCD-Anzeige
{
   port_out (lcdcom , 0x38); lcd_wait();
   port_out (lcdcom , 0x0c); lcd_wait();
   port_out (lcdcom , 0x06); lcd_wait();
   port_out (lcdcom , 0x01); lcd_wait();
}

void lcd_write ( unsigned char x, unsigned char y, char * text )
//   ---------
// Ausgabe an LCD-Anzeige
{  
   unsigned char offset;
   offset = 128 +  ( ( y%2 ) * 64 ) + ( x%32 ) ;
   port_out ( lcdcom, offset )     ; lcd_wait();
   while ( *text!='\0') { port_out ( lcddat, *text++); lcd_wait(); }
}


// Umwandlungs-Funktionen
// ----------------------

void fix ( unsigned int zahl, char buffaus[10], unsigned char lenmax )
//   ---
// wandelt die int zahl um in char buffaus[10] mit der Laenge lenmax
{  
   int j, laenge, lendif, akt;
   char buffein[10] ;
   itoa( zahl, buffein, 10 );
   laenge = strlen(buffein);  lendif = lenmax - laenge;
   j = 0;
   while ( j < lenmax )
      { akt = j - lendif;
      if ( akt < 0 ) { buffaus[j] = ' '          ; }
      else           { buffaus[j] = buffein[akt] ; }
   j++; }
   buffaus[lenmax] = '\0';
}

void fix0( unsigned int zahl, char buffaus[10], unsigned char lenmax )
//   ----
// wandelt die int zahl um in char buffaus[10] mit der Laenge lenmax
// mit fuehrenden Nullen
{  
   int j, laenge, lendif, akt;
   char buffein[10] ;
   itoa( zahl, buffein, 10 );
   laenge = strlen(buffein);  lendif = lenmax - laenge;
   j = 0;
   while ( j < lenmax )
      { akt = j - lendif;
      if ( akt < 0 ) { buffaus[j] = '0'          ; }
      else           { buffaus[j] = buffein[akt] ; }
   j++; }
   buffaus[lenmax] = '\0';
}


// ========================================================


void main (void)
//   -----------
// Hier wird durch Polling das Torbit des Funkuhrempfaengers abgefragt,
// waehrend gleichzeitig ein Timer die ablaufende Zeit in Millisekunden misst.
// Nachdem die jeweiligen Eingangssignalgrenzen festgestellt sind, wird anhand
// des Timerwertes ueberprueft, ob die Signalbreiten den geforderten 
// Zeitlaengen von 100 bzw. 200ms +/- Toleranz entsprechen. Ist dies nicht der
// Fall, kann nur ein schlechter Empfang vorliegen, dessen Wirkung durch eine
// Fehlermeldung angegeben wird. Nach einem vollstaendigen Empfang ohne Fehler,
// der minimal bis 2min benoetigt, wird die empfangene Zeit in die DOS-Uhr
// uebertragen.
{
   const    char timer    = 1;        // benutzter Timer
	    char pin      = 7;        // TorBit fuer DCF77-Impulse
	    char stoppin  = 2;        //        fuer STOP-Taste
//          int  portadr  = port2adr; // WEB1 linker  Port
	    int  portadr  = port1adr; // WEB1 rechter Port

	    char *line    = "\r\n------------";
	    char *pgmname = "dcf77-Funkuhr: ";

	    int  s        = 0;    // Sekundenzaehler nach Sync
	    int  bit      = 0;    // aktuelles Bit von dcf77
	    int  sync     = 0;    // synchronisiert nach Minutenmarke
	    int  pinmask  = 1;    // Tormaske fuer Uhranschluss
	    int  stopmask = 0;    //               Stop-Taster
	    int  stop        ;    // Stop-Signal
	    int  err0     = 0;    // Fehler im 0-Signal
	    int  err1     = 0;    // Fehler im 1-Signal
	    int  errcnt   = 0;    // Fehler-Zaehler
   unsigned      on       = 0;    // ms des 1 Signals
   unsigned      of       = 0;    // ms des 0 Signals
	    int  i, max;
   unsigned long    sum;
   unsigned char byte;
   char     zeile1[41], zeile2[41], buf1[41];

   init_msCount(timer);                   // ms-Zaehler init
// port_init();                           // SC12tor    init
					  // Probleme beim Einlesen von D7
   
   pinmask  = 1 << pin;                   // Maske fuer DCF77-Signal
   stopmask = 1 << stoppin;               // Maske fuer Stop-Taste

   // Programmstart
   printf("\r\n\n");
   printf("%s>> Start <<%s",pgmname, line);
   lcd_init();
   show_date();
   lcd_date();


do {  errcnt = 0;   // Anfang der Fehlerschleife

   while ( (port_in(portadr)&pinmask));   // Uhrstart mit HIGH
   while (!(port_in(portadr)&pinmask));
   sync   = 0; bit    = 0; s = 0; jahr = 0; sekunde = 1; sum = 1000;
   err1   = 0; errcnt = 0;
// maxmin = 0; maxmax = 0; minmin = 0; minmax = 0;

   start_timer(timer);

   while( !( sync==1 && s>=59 )) {        // Sek zaehlen, wenn sync
      on = 0; of = 0;

      while ( (port_in(portadr)&pinmask));     // dcf77 1-Signal warten
      on  = ms; ms = 0;                   // in msec
      bit = getDCF(on);                   // Signalkontrolle 0 oder 1
      if (sync) err1 = decodeDCF(s, bit); // Auswertung

      while (!(port_in(portadr)&pinmask));     // dcf77 0-Signal warten
      of  = ms; ms = 0;                   // in msec

      if ( of>1800 ) {                    // Minutenmarke ??
         if ( sync==0 ) {
            s = 0; sync = 1;              // synchronisiert
            printf("\n\n->Minutenbeginn nach %4u msec ", of);
            continue;   }
         else if ( sync==1 && s!=58 ) {   // zu lange Pause
            if ( err0 < 3 ) {
               s = 0; sync = 1; err0++;
               printf("\n neue Minute ? of: %u msec \n", of);
               continue;    }
            else break;
         }
      }

      // Kontrollausdruck auf COM
      if   (!sync) printf ("\n --.sec");
      else         printf ("\n %2u.sec", s);
      printf(" %3u + %4u = %4u msec", on, of, on+of);
      printf(" %2u...%3u %3u...%3u ", minmin, minmax, maxmin, maxmax);

      // Test auf Gesamtsummenzeiten
      sum = sum + on + of - 1000;
      printf("-> %4i msec", sum);

      printf(" bit: %2i", bit);
      if ( err1 != 0 ) {
	 printf(" Err: %2i", err1);
	 errcnt++; }

      // Kontrollausdruck auf LCD-Anzeige
      if (!sync) strcpy(zeile1," --");
         else { sprintf(zeile1," %2u", s);}
      sprintf(buf1,".sec %3u+%3u=%4u msec bit: %2d      ",
                         on, of, on+of,        bit);
      strcat(zeile1, buf1);
      lcd_write(0,1, zeile1);

   // sprintf(zeile2," min:%4d std:%4d ", minute, stunde);
   // lcd_write(0,1, zeile2);

      stop = ( port_in(portadr) & stopmask );
      if ( stop==4 ) {
         printf("\n\n Abbruch durch Stop-Taste \n");
         return; }
      s++;
   }

   printf("\n\n es gab %u Fehler:", errcnt);
   // Test auf Fehler: 0 -> Uhr setzen 1 -> von vorne
} while ( errcnt != 0 );   // Ende der Schleife

   // Abschluss des Programms
   stop_timer(timer);
   oldHandler(timer);

   if ( errcnt == 0 ) {
      printDCF();  set_time(); set_date();
      printf("\n Die Zeit wurde gesetzt!"); }
   show_date();
   lcd_date();
   lcd_write(0,1," Ende mit Stop-Taste                ");


   // weiter mit Uhrzeit-Anzeige im Sekundentakt bis Stop-Taste
   while( stop == 0)
   {
      sleep(1000);                               // CPU-Zeit schonender
      lcd_date();
      show_date();
      stop = ( port_in(portadr) & stopmask );
   }
   printf("\n\n %s >> Ende <<\r\n", pgmname);
   lcd_write(0,1," Ende                      ");
}
