// IPCDATE.CPP setzt Rechnerdatum auf PC oder IPC
//
// Author :   Diedrich Lamken Schaphuser Str. 2 28307 Bremen      12.01.2002
//
// Hardware : Mini-Webio von Elektor mit 2 Ports je 8 bit Adr. 100h und 101h
//            LCD-Anzeige 2 Zeilen je 40 Zeichen          Adr. 102h und 103h
//            Zusatzboard mit Stromversorgung 
//            und 2 AD-Wandler je 12 bit an Bit D0 und D1 Adr. 101h
//            DCF77-Modul                an Bit D7        Adr. 101h
//            Stop-Taste                 an Bit D2        Adr. 101h
//
// Software : Borland Turbo C++ V1.01 
//
// Quellen :  Elektor     Hefte 5 bis 9 / 2001
//            D.Schwarzer BBS Winsen/Luhe
//            C.Kuehnel   Embedded Webserver mit Single-Chip Controller

#pragma option -1               // 80186 Code erzeugen
 
// Includes
//---------
#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

// Definitionen
//------------- 
#define  HAL_INT  0xA1          // Interrupts
#define  PFE_INT  0xA2
#define  TCP_INT  0xAC
#define  DOS_INT  0x21

// globale Variablen
// -----------------

union    REGS  inregs;          // C-SoftInterrupt-Funktion
union    REGS  outregs;
struct   SREGS segregs;

unsigned char  tag, tagnr, monat;
         int   jahr;
	 char  *wotag[] = {"Sonntag"   ,"Montag" ,"Dienstag" ,"Mittwoch"
	                  ,"Donnerstag","Freitag","Sonnabend","Sonntag"};

//   Prototypes
//   ----------
void  get_date(void);
int   check_date(char *buf);
void  set_date(void);
void  usage(void);
double fmod(double x, double y);

void usage(void)
//   -----------
// Anzeige Bedienung  
{
   printf("\r\n Usage:                                 ");
   printf("\r\n Datum Abfrage : ipcdate <CR>           ");
   printf("\r\n Datum Eingabe : ipcdate tt.mm.jjjj<CR> ");
}

void get_date(void)
//   --------------
// holt das System-Datum des IPC
{
   inregs.h.ah = 0x2A;
   int86x(DOS_INT, &inregs, &outregs, &segregs);
   jahr  = outregs.x.cx;
   monat = outregs.h.dh;
   tag   = outregs.h.dl;
   tagnr = outregs.h.al;
   printf("\rAktuelles Datum: %s, der %0.2d.%0.2d.%4d", wotag[tagnr], tag, monat, jahr);
}

int  check_date(char *buf)
//   ---------------------
// prueft die Eingabe auf Datumformat tt.mm.jjjj
// bei Fehler :    return 0
// alles richtig : return 1
{
   char chr, tmp[5], valid;
   chr = '.';
   if ( buf[2]!=chr || buf[5]!=chr || buf[10]!=0 ) return (0);
   tmp[0] = buf[0]; tmp[1] = buf[1]; tmp[2] = 0; tag    = atoi(tmp);
   tmp[0] = buf[3]; tmp[1] = buf[4]; tmp[2] = 0; monat  = atoi(tmp);
   tmp[0] = buf[6]; tmp[1] = buf[7]; 
   tmp[2] = buf[8]; tmp[3] = buf[9]; tmp[4] = 0; jahr   = atoi(tmp);
   if ( monat>12 || jahr>2009 || jahr<2000 )             return (0);
   valid = 0;
   if ((monat==1 || monat==3  || monat==5  ||
        monat==7 || monat==8  || monat==10 ||
        monat==12 )                          && tag>0 && tag<32 ) valid = 1;
   if ((monat==4 || monat==6  || monat==9  ||
        monat==11 )                          && tag>0 && tag<31 ) valid = 1;
   if ((monat==2)                            && tag>0 && tag<29 ) valid = 1;
// if ( !fmod(jahr,4) && monat==2            && tag>0 && tag<30 ) valid = 1;
// if (  fmod(jahr,4) && monat==2            && tag>0 && tag<29 ) valid = 1;

   return (valid);                           // keine Fehler
}


void set_date (void)
//   ---------
// setzt die System-Datum des IPC
{
  printf("\rGesetztes Datum: %0.2d.%0.2d.%4d", tag, monat, jahr);

  inregs.h.ah = 0x2B;                        // SoftINT 21  Fkt 2D
  inregs.x.cx = jahr;
  inregs.h.dh = monat;
  inregs.h.dl = tag;
  int86x(DOS_INT, &inregs, &outregs, &segregs);
}

//**************************************************************
//   Main
void main (int argc, char *argv[])
//   -----------------------------
// die Charactereingabe auf dem IPC ueber COM funktioniert nicht
// deshalb neues Datum nur als Parameter
{
   char len,   *buf, *inbuf;
   int  valid;

   switch (argc)
   { case 1:                                  // keine Parameter
          get_date();                         // Datum holen
	  break;
     case 2:                                  // Parameter neues Datum
	  buf = argv[1];                      // Datum setzen
	  valid = check_date(buf);
	  if (valid) { set_date(); }
	  break;
     default: usage(); return ;
   }
}
