// IPCTIME.CPP setzt Rechnerzeit auf PC oder IPC
//
// Autor :    Diedrich Lamken Schaphuser Str. 2 28307 Bremen      08.01.2002
//
// Hardware : Mini-Webio von Elektor mit 2 Ports je 8 bit Adr. 100h und 101h
//            LCD-Anzeige 2 Zeilen je 40 Zeichen          Adr. 102h und 103h
//            Zusatzboard mit Stromversorgung 
//            und 2 AD-Wandler je 12 bit an Bit D0 und D1 Adr. 101h
//            Stop-Taste                 an Bit D2        Adr. 101h
//            DCF77-Modul                an Bit D7        Adr. 101h
//
// Software : Borland Turbo C++ V1.01 
//
// Quellen :  Elektor     Hefte 5 bis 9 / 2001
//            D.Schwarzer BBS Winsen/Luhe
//            C.Kuehnel   Embedded Webserver mit Single-Chip Controller

#pragma option -1               // 80186 Code erzeugen

// Includes
//---------
#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

// Definitionen
//------------- 
#define  HAL_INT  0xA1          // Interrupts
#define  PFE_INT  0xA2
#define  TCP_INT  0xAC
#define  DOS_INT  0x21

// globale Variablen
// -----------------

union    REGS  inregs;          // C-SoftInterrupt-Funktion
union    REGS  outregs;
struct   SREGS segregs;

unsigned char  stunde, minute, sekunde;

//   Prototypes
//   ----------
void  get_time(void);
int   check_time(char *buf);
void  set_time(void);

void get_time(void)
//   --------
// holt die System-Zeit des IPC
{
   inregs.h.ah = 0x2C;
   int86x(DOS_INT, &inregs, &outregs, &segregs);
   stunde      = outregs.h.ch;
   minute      = outregs.h.cl;
   sekunde     = outregs.h.dh;
   printf("\rAktuelle Uhrzeit:  %0.2d:%0.2d:%0.2d", stunde, minute, sekunde);
}

int  check_time(char *buf)
//   -------------
// prueft die Eingabe auf Zeitformat hh:mm:ss
// bei Fehler :    return 0
// alles richtig : return 1
{
   char chr, tmp[3];
   chr = ':';
   if ( buf[2]!=chr || buf[5]!=chr || buf[8]!=0 ) return (0);
   tmp[0] = buf[0]; tmp[1] = buf[1]; tmp[2] = 0; stunde   = atoi(tmp);
   tmp[0] = buf[3]; tmp[1] = buf[4]; tmp[2] = 0; minute   = atoi(tmp);
   tmp[0] = buf[6]; tmp[1] = buf[7]; tmp[2] = 0; sekunde  = atoi(tmp);
   if ( stunde>23 || minute>59 || sekunde>59 ) return (0);
   return (1);
}

void set_time (void)
//   ---------
// setzt die System-Zeit des IPC
{
  printf("\rGesetzte Uhrzeit:  %0.2d:%0.2d:%0.2d", stunde, minute, sekunde);

  inregs.h.ah = 0x2D;                        // SoftINT 21  Fkt 2D
  inregs.h.ch = stunde;
  inregs.h.cl = minute;
  sekunde++; if ( sekunde == 60 ) sekunde--; // seltsame Effekte
  inregs.h.dh = sekunde;
  int86x(DOS_INT, &inregs, &outregs, &segregs);
}

//**************************************************************
//   Main
void main (int argc, char *argv[])
//   -----------------------------
// die Charactereingabe auf dem IPC ueber COM funktioniert nicht
// deshalb neue Zeit nur als Parameter
{
   char len,   *buf, *inbuf;
   int  valid;

   switch (argc)
   { case 1:                                  // keine Parameter
	  get_time();                         // Zeit holen
	  break;
     case 2:                                  // Parameter neue Zeit
	  buf = argv[1];                      // Zeit setzen
	  valid = check_time(buf);
	  if (valid) { set_time(); }
	  break;
     default: return ; }
}
