// LCDZEIT.CPP zeigt Zeit auf WEBIO von Elektor
//            Test                                                03.03.2004
//
// Hardware : Mini-Webio von Elektor mit 2 Ports je 8 bit Adr. 100h und 101h
//            LCD-Anzeige 2 Zeilen je 40 Zeichen          Adr. 102h und 103h        
//            Zusatzboard mit Stromversorgung 
//            und 2 AD-Wandler je 12 bit an Bit D0 und D1 Adr. 101h
//            DCF77-Modul                an Bit D7        Adr. 101h
//            Stop-Taste                 an Bit D2        Adr. 101h
//
// Software : Borland Turbo C++ V1.01 
//
// Author :   Diedrich Lamken Schaphuser Str. 2 28307 Bremen      19.01.2002
//
// Quellen :  Elektor Hefte 5 bis 9 / 2001
//            D.Schwarzer BBS Winsen/Luhe
//            C.Kuehnel   Embedded Webserver mit Single-Chip Controller


#pragma option -1                // 80186 Code erzeugen

// Includes
//---------
#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

// Definitionen
//------------- 
#define  HAL_INT  0xA1           // Interrupts
#define  PFE_INT  0xA2
#define  TCP_INT  0xAC
#define  DOS_INT  0x21

#define  port1adr 0x100          // Port 1         
#define  port2adr 0x101          // Port 2
#define  lcdcom   0x102          // LCD-Anzeige com
#define  lcddat   0x103          //             Daten


// globale Variablen
// -----------------

union    REGS  inregs;          // C-SoftInterrupt-Funktion
union    REGS  outregs;
struct   SREGS segregs;

unsigned char  sekunde  = 0;    // nach dem Empfang 0
unsigned char  minute   = 0;    // ..von dcf77
unsigned char  stunde   = 0;    // ..von dcf77
unsigned char  tag      = 0;    // ..von dcf77
unsigned char  tagnr    = 0;    // ..von dcf77  0-So,1-Mo
unsigned char  monat    = 0;    // ..von dcf77
unsigned int   jahr     = 0;    // ..von dcf77
unsigned int   jahrt    = 2000;
	 char  *wotag[] =
         {"Sonntag",   "Montag", "Dienstag", "Mittwoch"
	 ,"Donnerstag","Freitag","Sonnabend","Sonntag"};


//   Prototypes
//   ----------
void get_time(void);
void lcd_write ( unsigned char x, unsigned char y, char *text );


void sleepw(int msec)
//   ------
// veranlasst eine Wartezeit, die in msec angegeben wird.
{
   inregs.h.ah = 0x09;                      // SoftINT: ACh, Fkt: 09h
   inregs.x.bx = msec;
   int86x(TCP_INT, &inregs, &outregs, &segregs);
}



unsigned char port_in ( int portadr )
//            -------
// Liest die Zustaende an den Eingabepins I0-I7
{
  inregs.h.ah = 0x80;                       // SoftINT A1  Fkt 80

  inregs.x.di = portadr;                    // an Adresse ausgeben                                        // geaendert Lamken
  inregs.x.bx = 0xFFFF;                     // wAND=0xFFFF
  inregs.x.cx = 0x0000;                     // wXOR=0x0000
  int86x(HAL_INT, &inregs, &outregs, &segregs);
  return outregs.h.al;
}

void port_out( int portadr, unsigned char ausgabe)
//   --------
// schreibt in die Ausgabepins O0-O7 den Wert 'ausgabe'
{
   inregs.h.ah = 0x81;                      // SoftINT: A1h, Fkt: 81h
   inregs.x.di = portadr;                   // Adresse ausgeben
   inregs.h.dh = 0;
   inregs.h.dl = ausgabe;
   inregs.x.bx = 0xFFFF;                    // wAND=0xFFFF
   inregs.x.cx = 0x0000;                    // wXOR=0x0000
   int86x(HAL_INT, &inregs, &outregs, &segregs);
}


// Funktionen der DOS-Uhr
// ----------------------

void set_time (void)
//   ---------
// nachdem die Zeit der Funkuhr empfangen wurde, uebertragen
// sie diese beiden Funktionen in den Zeitspeicher des SC12
// Betriebsystems. Hier wird die Zeit auf Quarzgenauigkeit
// fortgeschrieben (bis zu einem erneuten Empfang).
{
  inregs.h.ah = 0x2D;                       // SoftINT 21  Fkt 2D
  inregs.h.ch = stunde;
  inregs.h.cl = minute;
  inregs.h.dh = sekunde;
  int86x(DOS_INT, &inregs, &outregs, &segregs);
}


void get_time(void)
//   --------
// holt die System-Zeit des IPC
{
   inregs.h.ah = 0x2C;
   int86x(DOS_INT, &inregs, &outregs, &segregs);
   stunde      = outregs.h.ch;
   minute      = outregs.h.cl;
   sekunde     = outregs.h.dh;
// printf("Die aktuelle Zeit ist: %2d:%2d:%2d", stunde, minute, sekunde);
}


void set_date (void)
//   ----------
{
   inregs.h.ah = 0x2B;                       // SoftINT 21  Fkt 2B
   inregs.x.cx = jahr;
   inregs.h.dl = tag;
   inregs.h.dh = monat;
   int86x(DOS_INT, &inregs, &outregs, &segregs);
}


void get_date(void)
//   --------
// holt das System-Datum des IPC
{
   inregs.h.ah = 0x2A;
   int86x(DOS_INT, &inregs, &outregs, &segregs);
   jahr        = outregs.x.cx;
   monat       = outregs.h.dh;
   tag         = outregs.h.dl;
   tagnr       = outregs.h.al;
// printf("Das aktuelle Datum ist: %4d-%2d-%2d", jahr, monat, tag);
}

void show_date(void)
//   ---------
// gibt System-Datum und -Zeit des IPC aus
{
   get_date(); get_time();

   printf("\n %s, der %0.2d.%0.2d.%0.4d", wotag[tagnr], tag, monat, jahr);
   printf(" um %0.2d:%0.2d:%0.2d", stunde, minute, sekunde);
}

void lcd_date(void)
//   --------
// gibt System-Datum und -Zeit des IPC auf LCD-Anzeige aus
{
   char zeile1[41];

   get_date(); get_time();
   sprintf(zeile1," %s, der %2d.%0.2d.%4d um %2d:%0.2d:%0.2d",
      wotag[tagnr], tag, monat, jahr, stunde, minute, sekunde);
   lcd_write(0,0,zeile1);
}

// Funktionen der LCD-Anzeige
// --------------------------

void lcd_wait(void)
//   ----------
// wartet bis LCD-Anzeige fertig
{  do {}
   while (( volatile unsigned char) ( port_in(lcdcom) ) > 127 );
}

void lcd_init(void)
//   ----------
// initialisiert LCD-Anzeige
{
   port_out (lcdcom , 0x38); lcd_wait();
   port_out (lcdcom , 0x0c); lcd_wait();
   port_out (lcdcom , 0x06); lcd_wait();
   port_out (lcdcom , 0x01); lcd_wait();
}

void lcd_write ( unsigned char x, unsigned char y, char * text )
//   ---------
// Ausgabe an LCD-Anzeige
{  
   unsigned char offset;
   offset = 128 +  ( ( y%2 ) * 64 ) + ( x%32 ) ;
   port_out ( lcdcom, offset )     ; lcd_wait();
   while ( *text!='\0') { port_out ( lcddat, *text++); lcd_wait(); }
}


// ========================================================


void main (void)
//   -----------
// Anzeige der Zeit im Sekundentakt
{
   char stoppin  = 2;            //        fuer STOP-Taste
   int  portadr  = port1adr;
   int  stopmask = 0;            //               Stop-Taster
   int  stop     = 0;            // Stop-Signal

   stopmask = 1 << stoppin;      // Maske fuer Stop-Taste

   lcd_init();
   lcd_write(0,1," Ende mit Stop-Taste                ");
   show_date();

   // weiter mit Uhrzeit-Anzeige im Sekundentakt bis Stop-Taste
   while( stop == 0)
   {
      lcd_date();
//    show_date();
      sleepw(1000);
      stop = ( port_in(portadr) & stopmask );
   }
   printf("\n  >> Ende von LCDZEIT <<\n");
   lcd_write(0,1," Ende                      ");
}